#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "GLES2/gl2.h"
#include "EGL/egl.h"

#include "fractal.h"
#include "window.h"
#include "shader.h"
#include "esUtil.h"

static int xrot = 25, zrot = -30;
static float fractcenter[2] = {-1.786863383,-2.369371475e-06};
//static float fractcenter[2] = {-0.3912647562f, 0.6785442331f};
//static float fractcenter[2] = {0.0f, 0.0f};

static float minwidth = 4.775061513e-09;
static float fractwidth = 50.0;

static GLfloat ls_scale = 2.5;
	
static const unsigned int uiWidth  = 640;
static const unsigned int uiHeight = 480;

#ifdef _WIN32
static HWND hWindow;
static HDC  hDisplay;
#else
static Window hWindow;
static Display *hDisplay;
#endif

static EGLDisplay	sEGLDisplay;
static EGLContext	sEGLContext;
static EGLSurface	sEGLSurface;
static GLuint uiProgram, uiFragShader, uiVertShader;
static ESMatrix perspective, modelview, mvp;

static GLshort *landscape;

void init()
{
	// EGL configuration. Adopted from 'cube' application.
	EGLint aEGLAttributes[] = {
		EGL_RED_SIZE, 8,
		EGL_GREEN_SIZE, 8,
		EGL_BLUE_SIZE, 8,
		EGL_DEPTH_SIZE, 16,
		EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
		EGL_NONE
	};
	
	EGLint aEGLContextAttributes[] = {
		EGL_CONTEXT_CLIENT_VERSION, 2,
		EGL_NONE
	};
	
	EGLConfig	aEGLConfigs[1];
	EGLint		cEGLConfigs;
	
	GLint iLocPosition = 0;

	GLint iLocColour, iLocTexCoord, iLocNormal, iLocMVP;
	GLint iLocXangle, iLocYangle, iLocZangle;
	GLint iLocAspect, iLocLightPos, iLocSampler, iLocSampler2;

	GLenum myTex, myTex2;
	
	int i, linked;

	// EGL initialization. Adopted from 'cube' application.
#ifdef _WIN32
	hWindow = create_window(uiWidth, uiHeight);
#else
	hWindow = create_window("OpenGL ES 2.0 Example on a Linux Desktop", uiWidth,
		uiHeight, hDisplay, sEGLDisplay, aEGLConfigs[0], &colormap, &pVisual);
#endif

#ifdef _WIN32
	hDisplay = GetDC(hWindow);
#else
	hDisplay = XOpenDisplay(NULL);
#endif

	if (!hDisplay) {
		printf("Could not open display\n");
		exit(-1);
	}
	
	sEGLDisplay = eglGetDisplay(hDisplay);
	
	(eglInitialize(sEGLDisplay, NULL, NULL));
	eglChooseConfig(sEGLDisplay, aEGLAttributes, aEGLConfigs, 1, &cEGLConfigs);
	
	if (cEGLConfigs == 0) {
        printf("No EGL configurations were returned.\n");
		exit(-1);
	} 

	sEGLSurface = eglCreateWindowSurface(sEGLDisplay, aEGLConfigs[0], (EGLNativeWindowType)hWindow, NULL);

    if (sEGLSurface == EGL_NO_SURFACE) {
        printf("Failed to create EGL surface.\n");
		exit(-1);
    }

 	sEGLContext = eglCreateContext(sEGLDisplay, aEGLConfigs[0], EGL_NO_CONTEXT, aEGLContextAttributes);

    if (sEGLContext == EGL_NO_CONTEXT) {
        printf("Failed to create EGL context.\n");
		exit(-1);
    }

	eglMakeCurrent(sEGLDisplay, sEGLSurface, sEGLSurface, sEGLContext);

	// Shader initialization.
	process_shader(&uiVertShader, "shaders/shader.vert", GL_VERTEX_SHADER);
	process_shader(&uiFragShader, "shaders/shader.frag", GL_FRAGMENT_SHADER);

  uiProgram = glCreateProgram();

  // Attach shaders and link uiProgram.
  glAttachShader(uiProgram, uiVertShader);
  glAttachShader(uiProgram, uiFragShader);
  glLinkProgram(uiProgram);

	// Bind position to attribute 0.
	glBindAttribLocation (uiProgram, 0, "a_position");

	// Link the program.
	glUseProgram(uiProgram);

	// Check the link status.
	glGetProgramiv(uiProgram, GL_LINK_STATUS, &linked);
	if (!linked) {
    printf("Program not linked.\n");
		exit(-1);
  }

  // Enable depth testing.
	glEnable(GL_DEPTH_TEST);
}

void reshape(GLsizei w, GLsizei h)
{ 
	// Update perspective matrix.
	const float edge = NUMPOINTS / 2;
	esMatrixLoadIdentity(&perspective);
	esFrustum(&perspective, -edge, edge, -edge, edge, NUMPOINTS * 1.25, 3 * NUMPOINTS);
	glViewport(0, 0, w, h);
}

void display()
{
  int s;
  GLint mvpLoc;
  float left, bottom, step;
  unsigned int *leftp, *bottomp, *stepp;
  static int framec = 0;
  clock_t start;
  double duration;

	// Update modelview matrix.
	esMatrixLoadIdentity(&modelview);
	esTranslate(&modelview, 0, 0, -2.0 * NUMPOINTS);
	esRotate(&modelview, xrot, 1, 0, 0);
	esRotate(&modelview, zrot, 0, 0, 1);
	esScale(&modelview, ls_scale, ls_scale, 1);
	esTranslate(&modelview, - (NUMPOINTS-1) / 2, - (NUMPOINTS-1) / 2, 0);
	
	// Update MVP matrix by multiplying the modelview matrix with the perspective matrix.
	esMatrixMultiply(&mvp, &modelview, &perspective);

	// Clear buffers.
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepthf(200);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// Load the MVP matrix:
	mvpLoc = glGetUniformLocation(uiProgram, "u_mvpMatrix");
	glUniformMatrix4fv(mvpLoc, 1, GL_FALSE, (GLfloat*) &mvp.m[0][0]);

	glEnableVertexAttribArray(0);

	// Draw the landscape.
  left = fractcenter[0] - fractwidth / 2;
  bottom = fractcenter[1] - fractwidth / 2;
  step = fractwidth / (NUMPOINTS-1);
  landscape = initLandscape(left, bottom, step);
  //landscape = initLandscape(left, bottom, fractwidth);

  leftp = (unsigned int*) &left;
  bottomp = (unsigned int*) &bottom;
  stepp = (unsigned int*) &step;
  printf("left:%x, bottom:%x, step:%x\n", *leftp, *bottomp, *stepp);

  start = clock();
	for (s = 0; s < (NUMPOINTS-1); s++) {
    // For each triangle strip in the landscape.
    getNextLandscapeStrip(); // Generate the strip.
		glVertexAttribPointer(0, 3, GL_SHORT, GL_FALSE, 0, landscape);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, NUMPOINTS*2);
  }
  duration = (double)(clock() - start) / CLOCKS_PER_SEC;

  printf("Frame %d took %2.3f seconds to generate.\n", framec++, duration);

	glFlush();
  if (!eglSwapBuffers(sEGLDisplay, sEGLSurface)) {
    printf("Failed to swap buffers.\n");
  }
}

int main(int argc_in, char **argv) {
  int bDone = 0;
  MSG sMessage;

	init();
  reshape(uiWidth, uiHeight);
  display();

	// Enter event loop.
	while (!bDone) {
#ifdef _WIN32
		if(PeekMessage(&sMessage, NULL, 0, 0, PM_REMOVE)) {
			if(sMessage.message == WM_QUIT) {
				bDone = 1;
			} else {
				TranslateMessage(&sMessage);
				DispatchMessage(&sMessage);
			}
		}
#else
		while (XPending(hDisplay) > 0) {
      XNextEvent(hDisplay, &e);

      if (e.type == ButtonPress) {
        bDone = 1;
      }
    }
#endif

    /*
  	if ((ls_scale *= 1.02) >= 5.0) {
		  ls_scale = 2.5;
	  	if ((fractwidth *= 0.5) < minwidth) {
  			fractwidth = 50.0;
		  }
  	}
    */

    /*
  	if ((fractwidth *= 0.98) < minwidth) {
		  fractwidth = 50.0;
	  }
    display();
    */

    fractwidth *= 0.98;
    display();
    
    //Sleep(1000);
  }
  return 0;
}
