//
// fractal.c, by Per Christian Corneliussen, March 2011
//
// This is the software version of the fractal generator. It is layed out so
// that each C function corresponds to a hardware module. The algorithms should
// also be as similar as possible to how it is planned done in hardware.
//

#include <stdio.h>
#include <stdlib.h>
#include <math.h> // For hypotf()
#include "GLES2/gl2.h" // Emulator header, for GLshort.

#include "fractal.h" // Defines NUMPOINTS.

static GLshort *stripmem = NULL;
static GLshort *vertexmem = NULL;
static float left, bottom, step;
static int k; // Current strip number.

int getFractPoint(float re, float im)
// Get the number of iterations (n) before |z| exceeds 2. This function
// corresponds to the fractal point generator hardware module, mbrot.v.
{
	int n;
	float z_re = 0.0f, z_im = 0.0f;
	for (n = 1; n < 80; n++) {
		float z_re_old = z_re;
		z_re = z_re*z_re - z_im*z_im + re;
		z_im = 2.0f * z_re_old * z_im + im;
		
    // Simplified boundary check.
		if ((z_re >= 2.0f) || (z_re <= -2.0f) || (z_im >= 2.0f) || (z_im <= -2.0f))
			break;

    // Precice boundary check. This is the one used for the "comparison of
    // boundary check methods" chapter of the report.
    /* if (_hypotf(z_re, z_im) >= 2.0f)
      break; */
	}
	return n;
}

GLshort *initLandscape(float _left, float _bottom, float _step)
// Allocate memory for landscape generation.
{
  // Allocate memory for a single triangle strip. If initLandscape() has been
  // called previously, the old chunk of allocated memory is reused.
  if (vertexmem == NULL) {
    vertexmem = malloc(NUMPOINTS * 2 * 3 * sizeof(GLshort));
    stripmem = malloc(NUMPOINTS * 2 * sizeof(GLshort));
    if (stripmem == NULL) {
      fprintf(stderr, "Out of memory.\n");
      exit(-1);
    }
  }

  // Store configuration:
  left = _left;
  bottom = _bottom;
  step = _step;

  // Set current strip number to 0.
  k = 0;

  return vertexmem;
}

void getNextLandscapeStrip()
// Update memory area returned by initLandscape() with next triangle strip.
// initLandscape() must be called first for each frame. This module corresponds
// to the vertarray.v hardware module.
{
  int i = 0, j = 0;
  int x = 0, y = k;
  float re = left;
  float im_upper = bottom + step;// * (y+1);
  float im_lower = bottom;// + step * y;

  //printf("Strip no. %d:\n", k);

  // Generate fractal points for this strip. If on the first strip (k == 0), all
  // the points are generated. Otherwise, only the upper row is generated, and
  // the lower row is reused from the previous (lower) strip.
  while (i < (NUMPOINTS * 2)) {
    //re = left + step * (i/2);
    // Lower row:
    if (k == 0) stripmem[i++] = getFractPoint(re, im_lower);
    else        stripmem[i++] = stripmem[i+1];
    // Upper row:
    stripmem[i++] = getFractPoint(re, im_upper);
    re += step;
  }

  // Finally, the vertex array is generated. (This should be easily implemented
  // in hardware, as everything is constant except stripmem[].)

	for (i = 0; i < (NUMPOINTS*2); i++) {
		// For each vertex in the triangle strip.
    vertexmem[j++] = (y == k) ? x : x++;
    vertexmem[j++] = (y == k) ? y++ : y--;
    vertexmem[j++] = stripmem[i];
    //printf("%04x%04x%04x\n", vertexmem[j-3], vertexmem[j-2], vertexmem[j-1]);
  }

  // Finally, increment 'bottom' for next strip. Note how this whole thing is
  // made so that strips must be retreived in order.
  bottom += step;
  k++;
}

