//
// fpg.v, by Per Christian Corneliussen, March 2011.
//
// Fractal point generator. This module will calculate the height of a single
// point in the Mandelbrot set, specified with the (c_re, c_im) inputs. It will
// spend three clock cycles per iteration of the algorithm, and so the execution
// time is not constant. The 'valid' output will go high for one clock cycle
// when the algorithm is finished. A new operation may then be started
// immediately.
//
// There is no support for subnormal numbers. c_re is registered, but c_im is
// not. This is done to assist the vertex array generator (VAG.)
//

module fpg(
  input             clk,      // Clock.
  input             reset_n,  // Reset.
  input      [31:0] c_re,     // Real value of point to be calculated. FP32.
  input      [31:0] c_im,     // Corresponding imaginary value. FP32.
  input             enable,   // Synchronous start.
  input      [15:0] addr_in,  // Will be moved to addr_out whenever started.
  output reg [ 7:0] n,        // Number of iterations.
  output reg busy,            // Except after reset, busy='0' means n is valid.
                              // Note that after reset, n will be 0.
  output reg [15:0] addr_out
  );

  // Configurable parameters:
  parameter MAX_ITERATIONS = 80;

  // State encoding:
  parameter S_IDLE = 0, S_IT0 = 1, S_IT1 = 2, S_IT2 = 3;
  parameter M_OFF = 0, M_X2 = 1, M_NEGATE = 2;

  reg  [ 2:0] state;                      // State register.
  reg  [ 1:0] mulsel;                     // Specifies multiplicator post-op.
  reg  [31:0] mul_a, mul_b, add_a, add_b; // Functional unit inputs.
  wire [31:0] mul_out_t, add_out;         // Functional unit outputs.
  reg  [31:0] mul_out;                    // Multiplicator post-op output.
  wire [32:0] add_out_t;                  // Adder temporary output.
  reg  [31:0] z_re, z_im, c_re_r;         // Temporary storage.

  // Functional units. The FP32 adders uses a derived 33-bit format internally
  // that needs to be converted to and from FP32. Even though the adder
  // supports subnormal numbers, the multiplicator (and this module) does not,
  // thus the FP32<->FP33 conversion is done in a simplified way.
  vithar_lib_fp32_adder_main adder(
    .a({add_a[31:23], 1'b1, add_a[22:0]}),
    .b({add_b[31:23], 1'b1, add_b[22:0]}),
    .res(add_out_t)
  );  
  assign add_out = {add_out_t[32:24], add_out_t[22:0]};

  vithar_lib_f32_mul multiplier(
    .clk(clk),          // Note that clk and reset_n are unused unless specific
    .reset_n(reset_n),  // parameters are set. Otherwise the unit is fully
    .enable(1'b1),      // combinatorial and valid is connected to enable.
    .a(mul_a),
    .b(mul_b),
    .dout(mul_out_t),
    .valid()
  );

  always @ (mulsel, mul_out_t)
  begin
    // To support the data flow graph of the Mandelbrot algorithm, there is
    // logic for post-processing the output of the multiplication unit. Note
    // that the M_X2 (multiplication by 2) operation is made in the simplest
    // possible way - it does not support NaN or subnormal numbers.
    case (mulsel)
      M_X2: mul_out = {mul_out_t[31], mul_out_t[30:23] + 8'b1, mul_out_t[22:0]};  
      M_NEGATE: mul_out = {~mul_out_t[31], mul_out_t[30:0]};
      default: mul_out = mul_out_t;
    endcase
  end

  always @ (posedge clk or negedge reset_n)
  begin
    if (!reset_n) begin
      // Reset everything.
      mul_a <= 0;
      mul_b <= 0;
      add_a <= 0;
      add_b <= 0;
      z_re <= 0;
      z_im <= 0;
      c_re_r <= 0;
      addr_out <= 0;
      n <= 0;
      busy <= 0;
      state <= S_IDLE;
      mulsel <= M_OFF;

    end else begin
      // The Mandelbrot set is designed as an FSM with three repeating states
      // for each iteration of the algorithm.
      
      z_im <= add_out;  // z_im needs to be stored between S_IT1 and S_IT2;
                        // this can be implemented as add_out delayed.

      case (state)
        S_IDLE: begin
          // Unit idle. Send start values into the functional units.
          mul_a <= c_re;
          mul_b <= c_re;
          add_a <= 0;
          add_b <= c_im;
          z_re <= c_re;
          c_re_r <= c_re;
          mulsel <= M_OFF;

          if (enable) begin
            // If there is a pending request, initiate the algorithm by
            // jumping to the next state. 
            state <= S_IT0;
            busy <= 1'b1;
            n <= 0;
            addr_out <= addr_in;
          end else begin
            state <= S_IDLE;
            n <= n;
            busy <= 0;
          end
        end

        S_IT0: begin
          // Iteration cycle 0.
          mul_a <= add_out;
          mul_b <= add_out;
          add_a <= mul_out;
          add_b <= c_re_r;
          mulsel <= M_NEGATE;
          n <= n + 8'b1;

          // The boundary check is done here. This is done by checking whether
          // |z_re| or |z_im| exceeds 2, or whether n exceeds max. it. count.
          //$display("Boundary check at n=%d: z=(%x, %x)", n, z_re, add_out);
          if ((n == (MAX_ITERATIONS-1)) || (add_out[30:23] > 127) || (z_re[30:23] > 127))
          begin
            // Iteration count or z exceeds boundaries. Assert valid-output
            // and return to idle state.
            busy <= 0;
            state <= S_IDLE;
          end else begin
            busy <= 1'b1;
            state <= S_IT1;
          end
        end

        S_IT1: begin
          // Iteration cycle 1.
          mul_a <= z_re;
          mul_b <= z_im;
          add_a <= mul_out;
          add_b <= add_out;
          z_re <= z_re;
          mulsel <= M_X2;
          n <= n;
          busy <= 1'b1;
          state <= S_IT2;
        end

        S_IT2: begin
          // Iteration cycle 2.
          mul_a <= add_out;
          mul_b <= add_out;
          add_a <= mul_out;
          add_b <= c_im;
          z_re <= add_out;
          mulsel <= M_OFF;
          n <= n;
          busy <= 1'b1;
          state <= S_IT0;
        end

        default: begin
          // Undefined state.
          mul_a <= 32'bx;
          mul_b <= 32'bx;
          add_a <= 32'bx;
          add_b <= 32'bx;
          z_re <= 32'bx;
          mulsel <= M_OFF;
          n <= 0;
          busy <= 0;
          state <= S_IDLE;
        end
      endcase
    end
  end
endmodule
