//
// vag.v, by Per Christian Corneliussen, May 2011.
//
// Vertex array generator. This module will generate a vertex array
// representing one triangle strip of the fractal landscape.
// The generated array itself contains fixed-point 16-bit (GLshort) values
// only. However, FP32 coordinates specifying a location in the Mandelbrot
// set (bottom left coordinates) and a zoom level (given as a step size) must
// be provided. See port list.
// 

module vag(
  input           clk,      // Clock.
  input           reset_n,  // Reset.
  input    [31:0] left,     // Real value of the bottom left fractal coord.
  input    [31:0] bottom,   // Corresponding imaginary value.
  input    [31:0] step,     // Distance between fractal points.
  input           start,    // Initiate generation of new landscape.
  input           next,     // Initiate generation of next triangle strip.
  output reg      ready,    // Output data ready. After 'ready' toggles, a new
  output   [47:0] arrout    // set of vertex coordinates will be available each
                            // clock cycle on 'arrout'.
  );

  // Configurable parameters:
  parameter NUMPOINTS = 400;  // Number of points in each direction. The number of
                            // vertices per triangle strip will be 2*NUMPOINTS.
  parameter NUMUNITS = 2;   // Number of fractal point generators.

  // State encoding:
  parameter S_IDLE = 0, S_FILL = 1, S_FLUSH = 2, S_READOUT = 3;

  //-----------
  // Variables:
  //-----------
  
  // Adders:
  reg  [31:0] add_a;
  wire [31:0] add_b;
  wire [31:0] add_out;
  wire [32:0] add_out_t;
  wire  [15:0] x_inc, y_inc;

  // Control state machine (CSM):
  reg [31:0] re, im;        // The current fractal coordinates. Connected to
                            // c_re and c_im inputs on all FPGs.
  reg [31:0] re_reg, im_reg;// Registrered fractal coordinates.
  reg [15:0] x, x_reg;      // The current column. (x vertex coordinate)
  reg [15:0] y, y_reg;      // The current row. (y vertex coordinate)
  reg firststrip;           // Whether this is the first strip or not.
  reg upper;                // Whether sending the upper or lower vertex.
  reg [1:0] state;          // Current state.

  // FPGs:
  reg [0:NUMUNITS-1] fpg_en;
  wire [0:NUMUNITS-1] fpg_busy;
  wire [7:0] fpg_n [0:NUMUNITS-1];
  wire [15:0] fpg_addr_out [0:NUMUNITS-1];

  // FPG arbiter: 
  reg arbt_enable;  // Whether to process the fractal point or not.
  reg arbt_ready;   // Will go high for one cc when the fractal point specified
                    // by the above variables is received and being processed.
  wire arbt_flushed;// Will stay high when no fractal points are being processed.
  reg [0:NUMUNITS-1] arbt_fpg_wb, arbt_fpg_wb_next; // Whether an FPG had its
                                                    // result stored yet.
  integer i;        // Loop counter.

  // Z memory:
  reg [7:0] zmem0 [0:NUMPOINTS-1]; // Ram for one row of z coordinates.
  reg [7:0] zmem1 [0:NUMPOINTS-1]; // Ram for one row of z coordinates.
  reg zmem_w_en;
  reg zmem_bufno;
  reg [15:0] zmem_waddr;
  reg [7:0] zmem_wdata;
  wire [15:0] zmem_raddr;
  wire [7:0] zmem_rdata;

  //--------
  // Adders:
  //--------
  
  // Floating point (FP32) adder. The FP32<->FP33 conversion is done in a
  // simplified way as in the FPGs. (I.e., dropped support for subnormal numbers.)
  vithar_lib_fp32_adder_main adder(
    .a({add_a[31:23], 1'b1, add_a[22:0]}),
    .b({add_b[31:23], 1'b1, add_b[22:0]}),
    .res(add_out_t)
  );
  assign add_out = {add_out_t[32:24], add_out_t[22:0]};

  // 16-bit integer incrementors.
  assign x_inc = x_reg + 16'b1;
  assign y_inc = y_reg + 16'b1;

  //-----------------------------
  // Control state machine: (CSM)
  //-----------------------------
 
  assign add_b = step;      // The second add operand is always 'step'.
  assign arrout[47:32] = x_reg; // x coordinate.
  assign arrout[31:16] = (upper ? y_inc : y_reg); // y coordinate.
  assign arrout[15: 0] = zmem_rdata; // z coordinate.
  assign zmem_raddr = x_reg;

  always @ (*)
  // Combinatorial block for controlling feeding of fractal points to FPGs.
  begin
    // Default values:
    x <= x_reg;
    y <= y_reg;
    add_a <= 32'bx;
    re <= 32'bx;
    im <= im_reg;
    arbt_enable <= 0;

    case (state)
      S_IDLE: begin
        x <= 0;
        if (start) begin
          // If starting on a new frame, send bottom left coordinates to FPGs.
          y <= 0;
          re <= left;
          im <= bottom;
          arbt_enable <= 1'b1;
        end else if (next) begin
          // If starting on a new strip, increment im with 'step'.
          y <= y_inc;
          add_a <= im_reg;
          re <= left;
          im <= add_out; // im += step
          arbt_enable <= 1'b1;
        end
      end
      S_FILL: begin
        x <= x_inc; // x++ 
        add_a <= re_reg;
        im <= im_reg; // Within a row, the imaginary value (im) will remain constant.
        arbt_enable <= 1'b1;

        if (arbt_ready) begin
          // The previous fractal point sent to the FPGs was consumed. Calculate
          // next fractal point.
          x <= x_inc; // x++
          re <= add_out; // re += step
        end else begin
          x <= x_reg;
          re <= re_reg; // Keep old value, waiting for a free FPG.
        end
      end
      S_FLUSH: begin
        x <= 0;
        if (arbt_flushed && firststrip) begin
          // If the FPGs have been flushed and this is the first triangle strip,
          // start calculating the upper row.
          add_a <= im_reg;
          re <= left;
          im <= add_out;
          arbt_enable <= 1'b1;
        end
      end
    endcase
  end

  always @ (posedge clk or negedge reset_n)
  // State control.
  begin
    if (!reset_n) begin
      state <= S_IDLE;
      re_reg <= 0;
      im_reg <= 0;
      x_reg <= 0;
      y_reg <= 0;
      firststrip <= 0;
      ready <= 0;
      upper <= 0;
      zmem_bufno <= 0;
    end else begin
      re_reg <= re;
      im_reg <= im;
      x_reg <= x;
      y_reg <= y;

      // Default values:
      ready <= 0;
      upper <= 0;

      case (state)
        S_IDLE: begin
          // Wait for either 'start' or 'next'.
          if (start) begin
            state <= S_FILL;
            firststrip <= 1'b1;
            zmem_bufno <= 0; // Always start with the first buffer.
          end else if (next) begin
            state <= S_FILL;
            firststrip <= 0;
            zmem_bufno <= ~zmem_bufno; // Change to the other buffer.
          end else begin
            state <= S_IDLE;
          end
        end
        S_FILL: begin
          // Calculate the z coordinates for the current vertex row.
          if (x == (NUMPOINTS-1)) begin
            // If at the right end of the row, jump to flush state.
            state <= S_FLUSH;
          end else begin
            state <= S_FILL;
          end
        end
        S_FLUSH: begin
          // Wait for final fractal points to be calculated.
          if (arbt_flushed) begin
            // Done. If started with 'next', jump to the readout state.
            // Otherwise if started with 'start', change the memory buffer
            // and continue with the upper row of the triangle strip.
            if (firststrip) begin
              state <= S_FILL;
              firststrip <= 0;
              zmem_bufno <= ~zmem_bufno;
            end else begin
              state <= S_READOUT;
              ready <= 1'b1;
            end
          end
        end
        S_READOUT: begin
          // Vertex array generated. Transmit values out of module.
          x_reg <= (upper ? x_inc : x);
          upper <= ~upper;

          if (!upper && (x == (NUMPOINTS-1))) begin
            // If the vertex following this clock cycle is the final one, jump
            // back to the idle state.
            state <= S_IDLE;
          end
        end
      endcase
    end
  end

  //------
  // FPGs:
  //------
  // The fractal point generators (FPGs) calculate the iteration counts of
  // incoming fractal points. (I.e., the z coordinates of the vertices.)
 
  generate
    // Generate FPGs.
    genvar gi;
    for (gi=0; gi < NUMUNITS; gi=gi+1) begin : UNITS
      fpg fpgi (
        .clk(clk), .reset_n(reset_n),
        .c_re(re), .c_im(im), .enable(fpg_en[gi]), .addr_in(x),
        .n(fpg_n[gi]), .busy(fpg_busy[gi]), .addr_out(fpg_addr_out[gi])
      );
    end
  endgenerate

  //-------------
  // FPG arbiter:
  //-------------
  // The purpose of the FPG arbiter is to keep track of free FPGs and store
  // results in the z memory. The variables below are used to interface the
  // arbiter by the control state machine (CSM.)
  
  reg arbt_break1, arbt_break2; // Temporary variables.

  assign arbt_flushed = & arbt_fpg_wb;

  always @ (*)
  begin
    // If any of the FPGs are done, write the result to the z memory. If several
    // FPGs are done, only the first one will have its result written this cc.
    zmem_w_en <= 0;
    zmem_waddr <= 0;
    zmem_wdata <= 0;
    arbt_ready <= 0;
    arbt_fpg_wb_next <= arbt_fpg_wb;
    fpg_en <= 0;
    arbt_break1 = 0;
    arbt_break2 = 0;
    for (i=0; i < NUMUNITS; i=i+1) begin
      // For each FPG.
      if (!fpg_busy[i]) begin
        // This FPG is free.
        if (!arbt_fpg_wb[i] && !arbt_break1) begin
          // The result of the previous operation is not yet written to the
          // z memory. Do it.
          zmem_w_en <= 1'b1;
          zmem_waddr <= fpg_addr_out[i];
          zmem_wdata <= fpg_n[i];
          arbt_fpg_wb_next[i] <= 1'b1;
          arbt_break1 = 1'b1;
        end
        // Also assign any pending request from the CSM to this FPG.
        if (arbt_enable && !arbt_break2) begin
          arbt_ready <= 1'b1;
          fpg_en <= 0;
          fpg_en[i] <= 1'b1;
          arbt_fpg_wb_next[i] <= 0;
          arbt_break2 = 1'b1;
        end
      end
    end
  end

  always @ (posedge clk or negedge reset_n)
  begin
    if (!reset_n) begin
      arbt_fpg_wb <= {NUMUNITS{1'b1}};
    end else begin
      if (arbt_fpg_wb != arbt_fpg_wb_next)
      arbt_fpg_wb <= arbt_fpg_wb_next;
    end
  end

  //----------
  // Z memory:
  //----------

  // Readout from the memories. 'zmem_bufno' is the buffer used for the upper row
  // of vertices. 'upper' is whether to read a lower or upper row vertex.
  assign zmem_rdata = upper ?
                          (zmem_bufno ? zmem1[zmem_raddr] : zmem0[zmem_raddr]) :
                          (zmem_bufno ? zmem0[zmem_raddr] : zmem1[zmem_raddr]);

  always @ (posedge clk)
  begin
    if (zmem_w_en) begin
      if (zmem_bufno)
        zmem1[zmem_waddr] <= zmem_wdata;
      else
        zmem0[zmem_waddr] <= zmem_wdata;
    end
  end
endmodule
